;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Effect Name : Band Pass Filter
;
; Description : 2nd order iir band pass filter
;
; Nb In       : 1
; Nb Out      : 1
;
; Input       : mono sound signal
; Output      : filtered signal
;
; Cycles      : 67 + 12 cycles (-1=unknownw)
; Memory      : 12 words (without init)
; Delay Line  : no
;
; Author      : Gregor
; Date        : 08/08/2003
; Version     : 1.1
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

ABS	INIT	0x3FF
	
	;; SEC1 Control parameters
ABS	FRed 0x0		; resonance frequency = FRes/(FSamp/2)
ABS	W    0x1		; band-width
	
	;; SEC2 Initial values of control parameters (S3.24 format)
IFRed	EQU 0x0100000
IW	EQU 0x1800000
	
	;; SEC3 Constants definitions
C72     EQU -0.0139396
C56     EQU -0.0169386
C42     EQU -0.0238671
C30     EQU -0.0332733
C20     EQU -0.0500052
C12     EQU -0.0833286
C6      EQU -0.166667
C2      EQU -0.499999
CPi	EQU 3.1415926

	;; SEC4 Memory variables definition
ABS	A1	0x2
ABS	A2	0x3
ABS	B0	0x4
ABS	B1	0x5
ABS	B2	0x6
ABS	K	0x7
	
ABS	X1	0x8		; x_{i-1}
ABS	X2	0x9		; x_{i-2}
ABS	Y1	0xa		; y_{i-1}
ABS	Y2	0xb		; y_{i-2}

	;; SEC5 Initial values of memory variables

	;; SEC6 Initialization
CM	0x40000	INIT
SKIP	!Z	EndInit
C	IFRed
SCA	0x0 FRed
C	IW
SCA	0x0 W
SCA	0x0 X1
SCA	0x0 X2
SCA	0x0 Y1
SCA	0x0 Y2
C       0x1000000     ; 1->A
SCA     0x0     INIT        ; do not reinitialize next time
EndInit:

	;; SEC7 Module code

	;;  STEP 1: calculates coefficients
CM	CPi FRed		; load Pi*FRed
AAC	0x0			; load (pi*FRed)^2 in A
X1AC	0x0			; store it in B
CAD	C56 0x100		; 1-x^2/56 -> A
BAC	0x0
CAD	C30 0x100		; 1-x^2/30 (1-x^2/56) -> A
BAC	0x0
CAD	C12 0x100		; 1-x^2/12 (...) -> A
BAC	0x0
CAD	C2 0x100		; cos = 1-x^2/2 (...) -> A
SCB	0x40000 A1		; cos(pi*FRed) in A1
CAD	C72 0x100		; 1-x^2/72 -> A
BAC	0x0
CAD	C42 0x100		; 1-x^2/42 (1-x^2/72) -> A
BAC	0x0
CAD	C20 0x100		; 1-x^2/20 (...) -> A
BAC	0x0
CAD	C6 0x100		; 1-x^2/6 (...) -> A
SXCA	0x0 K			; K = sin(pi*FRed)/(pi*FRed) in K and B.

	;; calculates sinh(...)
LOGB
CAD	0x3c0000 0x0
X1AC	0x0
EXPB				; 1/K in A
AMC	0x0 W
DAC	0x20 0x0		; W/8K in A
SXCA	0x3c0000 B0		; W/8K in B, -W/8K in A, B0 irrelevant
SEXPB	B0			; use B0 to store -W/8K, 16^(W/8K) -> A
LCA	0x20000 B0		; 1/2*16^(W/8K) -> A, -W/8K -> B
SEXPB	B0			; use B0 to store 1/2*16^(W/8K), 16^(-W/8K) -> A
CAM	0x3e0000 B0		; sinh in A
AMC	0x0 FRed
CAD	CPi 0x0
AMC	0x0 K			; ALPHA value in A
SCA	0x40000 B0		; B0 = alpha (! not anymore the temp variable !)
1AC	0xF000000
SCA	0x40000 A2		; A2 = ALPHA-1
1AC	0x2000000		; 1+ALPHA -> A
X1AC	0x0
LOGB
CAD	0x3c0000 0x0
X1AC	0x0
EXPB
X1AC	0x0			; 1/(1+ALPHA) -> A, B
AMC	0x0 A2			; rescale A2
SCA	0x0 A2			; and store it
BMC	0x0 B0			; rescale B0
SCA	0x3c0000 B0		; store it
SCA	0x0 B2			; and fill B2
CM	0x80000 A1
BAC	0x0			; update A1 (*2B)
SCA	0x0 A1
SCA	0x0 B1

	;;  STEP 2: filter
CM	0x40000 X2		; x(i-2) -> A
AMC	0x0 B2			; a2 x(i-2) -> A
XCM	0x40000 X1		; A -> B and x(i-1) -> A
SCA	0x40000 X2		; A unchanged, X1 will be X2 at next run
AMB	B1			; a1 x(i-1) + a2 x(i-2) -> A
XCM	0x40000 IN1		; A -> B and x(i) -> A
SCA	0x40000 X1		; A unchanged, IN will be X1 at next run
AMB	B0			; a0 x(i)+a1 x(i-1)+a2 x(i-2) -> A
XCM	0x40000 Y2		; A -> B and y(i-2) -> A
AMB	A2			; b2 y(i-2)+a0 x(i)+a1 x(i-1)+a2 x(i-2) -> A
XCM	0x40000 Y1		; A -> B and x(i-1) -> A
SCA	0x40000 Y2		; A unchanged, Y1 will be Y2 at next run
AMB	A1			; A not contins full output !
SCA	0x40000 Y1		; A unchanged, OUT will be Y1 at next run
SCA	0x0 OUT1		; write output !
