////////////////////////////////////////////////////////////////////////////
// Name:        fileio.cpp
// Project:     GUI for Axoris AL3101 application generator
// Author:      Gregor(y) Soyez
// Version:     1.1
// Created:     22/07/02
// Copyright:   (c) Axoris
/////////////////////////////////////////////////////////////////////////////

// local header
#include "fileio.h"
#include "console.h"
#include "gldefs.h"
#include "Config.h"
#include "mpproject.h"

// wxWindows headers
#include <wx/wx.h>

// ctor & dtor
CFile::CFile(CglModuleList *i_mods, CLinkList *i_links, int i_offset, wxFrame *i_MainFrame){
#ifdef DEBUG
  fprintf(stderr, "DBG: fileio: Entering constructor\n");
#endif
  
  mods  = i_mods;
  links = i_links;
  offset = i_offset;

  strcpy(Name, "");
  NewFile=1;
  IsModified=0;

  MainFrame = i_MainFrame;
}

CFile::~CFile(){
#ifdef DEBUG
  fprintf(stderr, "DBG: fileio: Entering destructor\n");
#endif
  OnQuit();
}

int CFile::SetOffset(int i_offset){
  offset = i_offset;
  return 0;
}

void CFile::OnQuit(){
  int ToSave;
  if ((IsModified) || (mods->IsModified))
    ToSave = wxMessageBox("Save actual project ?", "Save?", wxYES_NO);
  else
    ToSave=wxNO;
  if (ToSave == wxYES)
    Save();
}

// prepare a new project
int CFile::New(){
#ifdef DEBUG
  fprintf(stderr, "DBG: fileio: Entering new\n");
#endif

  // ask if we need to save the current project
  int ToSave;

  if ((IsModified) || (mods->IsModified))
    ToSave = wxMessageBox("Save actual project ?", "Save?", wxYES_NO);
  else
    ToSave=wxNO;
  if (ToSave == wxYES)
    Save();

  // create new project
  return PerformNew();
}

int CFile::PerformNew(){
#ifdef DEBUG
  fprintf(stderr, "DBG: fileio: Entering PerformNew\n");
#endif

  // reset the project name
  strcpy(Name, "");
  strcpy(info.Name, "?");
  strcpy(info.Descr, "?");
  strcpy(info.InDescr, "?");
  strcpy(info.OutDescr, "?");
  info.NbIn = info.NbOut = 1;
  info.Cycles = info.MemUsage = info.UseDelay = 0;
  info.InitCycles = 4;

  NewFile=1;

  // indicate that we start a new project.
  //(*AppConsole) << "Starting a new project.\n";
  wxLogMessage("Starting a new project.");

  // reset links and modules lists
  links->Reset();
  mods->Reset();

  // create 'input' and 'output' modules
  mods->Add("Input", "", "Input", 10, 2);
  mods->Add("Output", "", "Output", 10, 17);

  IsModified=1;

  wxString title;
  title << "AgALag development interface"; 
  MainFrame->SetTitle(title);
  return 0;
}

// open an existing project
int CFile::Open(){
#ifdef DEBUG
  fprintf(stderr, "DBG: fileio: Entering Open\n");
#endif

  // ask if we need to save the current project
  int ToSave;
  if ((IsModified) || (mods->IsModified))
    ToSave = wxMessageBox("Save actual project ?", "Save?", wxYES_NO);
  else
    ToSave=wxNO;
  if (ToSave == wxYES)
    Save();

  // retreive the project to open
  CConfig *Config;
  Config=new CConfig;
  wxString directory = Config->ProjectDir;
  delete Config;
  wxString s=wxFileSelector("Choose the project name", directory, "", "", "*.axp", wxOPEN);
  if (!(s)){
    return 1;
  }

  strcpy(Name, s.c_str());
  NewFile=0;
  wxString lm;
  lm << "Opening project '" << Name << "'.";
  wxLogMessage(lm);

  // open the requested project
  return Open(Name);
}


int CFile::Open(char *FName){
#ifdef DEBUG
  fprintf(stderr, "DBG: fileio: Entering Open(2)\n");
#endif

  FILE *flux=NULL;
  int n;

  // check if name is not "" (Required for Win32)
  if ((FName==NULL) || (strcmp(FName, "")==0)){
    wxString lm;
    lm << "'"<< FName << "' cannot be opened, creating a new file.\n";
    wxLogMessage(lm);
    strcpy(FName, "");
    NewFile=1;
    return PerformNew();
  }

  // check if the provided project exists
  flux=fopen(FName, "r");
  if (flux==NULL){
    wxString lm;
    lm << "'"<< FName << "' cannot be opened, creating a new file.\n";
    wxLogMessage(lm);
    strcpy(FName, "");
    NewFile=1;
    return PerformNew();
  } else {
    NewFile=0;
  }

  // save project name
  strcpy(Name, FName);

  // reset links and modules lists
  links->Reset();
  mods->Reset();

  // read file information
  if (ReadProjectInfo(flux)!=0){
    wxString lm2;
    lm2 << "'"<< FName << "' has incorrect format, creating a new file.\n";
    wxLogMessage(lm2);
    strcpy(FName, "");
    NewFile=1;
    return PerformNew();
  }
  n=mods->ReadModuleSection(flux, offset);
  links->ReadLinkSection(flux, mods->List,n);
  mods->ReadPropertiesSection(flux);

  // close the project file
  fclose(flux);

  // try to load a Miss PArker project
  LoadMPProject(Name, mods->List);

  // new project is unmodified !
  IsModified=0;
  mods->IsModified=0;
  
  wxString title;
  title << "AgALag development interface: " << Name; 
  MainFrame->SetTitle(title);

  return 0;
}

// save current project
int CFile::Save(){
#ifdef DEBUG
  fprintf(stderr, "DBG: fileio: Entering Save\n");
#endif

  FILE *script_flux;
  
  if (NewFile)
    return SaveAs();

  wxString lm;
  lm << "Saving project '" << Name << "'.\n";
  wxLogMessage(lm);

  script_flux=fopen(Name, "w+");
  if (script_flux==NULL){
    wxLogMessage("Project file creation failed.\n");
    return 1;
  }
  SaveProjectInfo(script_flux);
  // create module section
  mods->CreateModuleSection(script_flux);
  links->CreateLinkSection(script_flux);
  mods->CreatePropertiesSection(script_flux);
  fclose(script_flux);

  IsModified=0;
  mods->IsModified=0;

  wxString title;
  title << "AgALag development interface: " << Name; 
  MainFrame->SetTitle(title);
  return 0;
}

// save current project with another file name
int CFile::SaveAs(){
#ifdef DEBUG
  fprintf(stderr, "DBG: fileio: Entering SaveAs\n");
#endif

  // retreive filename to save
  // wxOVERWRITE_PROMPT used ... can remove it if you want !
  CConfig *Config;
  Config=new CConfig;
  wxString directory = Config->ProjectDir;
  delete Config;
  wxString s=wxFileSelector("Choose the project name", directory, "", "", "*.axp", wxSAVE|wxOVERWRITE_PROMPT);

  if (!s.IsNull()){
    strcpy(Name, s.c_str());
    NewFile=0;
    return Save();
  }

  IsModified=0;
  mods->IsModified=0;
  return 1;
}

int CFile::ReadProjectInfo(FILE *flux){
  int len, tmp;
  char *rc;

  // check if the axp file contains descriptions 
  // i.e. the following informations:
  //   <project name>
  //   <description>
  //   <Nb In>
  //   <Nb Out>
  //   <input description>
  //   <output description>
  rc = fgets(info.Name, 255, flux);
  if (rc==NULL) return 1;
  rc = fgets(info.Descr, 255, flux);
  if (rc==NULL) return 1;
  tmp = fscanf(flux, "%d\n", &(info.NbIn));
  if ((tmp==0) || (tmp==EOF)) return 1;
  tmp = fscanf(flux, "%d\n", &(info.NbOut));
  if ((tmp==0) || (tmp==EOF)) return 1;
  rc = fgets(info.InDescr, 255, flux);
  if (rc==NULL) return 1;
  rc = fgets(info.OutDescr, 255, flux);
  if (rc==NULL) return 1;

  // removes linefeeds
  len = strlen(info.Name)-1;
  while ((info.Name[len]==0x0A) || (info.Name[len]==0x0D)){ info.Name[len]='\0'; len--;};
  
  len = strlen(info.Descr)-1;
  while ((info.Descr[len]==0x0A) || (info.Descr[len]==0x0D)){ info.Descr[len]='\0'; len--;};
  
  len = strlen(info.InDescr)-1;
  while ((info.InDescr[len]==0x0A) || (info.InDescr[len]==0x0D)){ info.InDescr[len]='\0'; len--;};
  
  len = strlen(info.OutDescr)-1;
  while ((info.OutDescr[len]==0x0A) || (info.OutDescr[len]==0x0D)){ info.OutDescr[len]='\0'; len--;};

  return 0;
}

int CFile::SaveProjectInfo(FILE *flux){
  fprintf(flux, "%s\n", info.Name);
  fprintf(flux, "%s\n", info.Descr);
  fprintf(flux, "%d\n", info.NbIn);
  fprintf(flux, "%d\n", info.NbOut);
  fprintf(flux, "%s\n", info.InDescr);
  fprintf(flux, "%s\n", info.OutDescr);

  return 0;
}
