 //--------------------------------------------------------------------//
 //  AgALag : Axoris group AL3101 application generator                //
 //                                                                    //
 //                 _         _    _      _                            //
 //                / \____   / \  | |    / \____                       //
 //               / _ \  _\ / _ \ | |   / _ \  _\                      //
 //              / /_\ \  __ /_\ \| |  / /_\ \  __                     //
 //             / _____ \ \ \____ \ |_/______ \ \ \                    //
 //            /_/     \_\___\   \_\____|    \_\___\                   //
 //                                                                    //
 //  File: modorder.cpp                                                //
 //  Description: classes dealing with module structure linearization  //
 //  Version: 1.0                                                      //
 //  Author: Gregor                                                    //
 //  e-mail: Gregor.Soyez@swing.be                                     //
 //                                                                    //
 //  Copyright (C) 2002 Axoris                                         //
 //--------------------------------------------------------------------//
 #include <string.h>
 #include <stdio.h>
 #include <ctype.h>

 #include "modorder.h"
 #include "messages.h"

 //----------------------------------------------------------------------------
 // CLinkList class implementation
 // Class dealing with the TLinkList stack. It should be used by each
 // module output channel to specify which modules it has to be connected to.
 // It is implemented as a LIFO.
 //-----------------------------------------------------------------------------

 // ctor & dtor
 CLinkList::CLinkList(){
   LinkList=NULL;
   Current=NULL;
   NbLinks=0;
 }

 CLinkList::~CLinkList(){
   while (LinkList!=NULL){
     Current=(TLinkList*) LinkList->Next;
     delete LinkList;
     LinkList=Current;
   };
 }

 // append a link to the stack. The link is given by a pointer to the 
 // destination module (as an index and a void*) and its input channel
 // This returns MODORDER_OK              if successfull
 //              MODORDER_LINK_EXISTS     if link already exists
 //              MODORDER_INVALID_CHANNEL if channel is not valid
 //              MODORDER_CHANNEL_BUSY    if channel is already used
 int CLinkList::Append(int ModuleIndex, void *Module, int InputChannel, int State){
   TLinkList *AddedLink;

   if (((CModule*) Module)->IsUsed(InputChannel)==MODORDER_USED)
     return MODORDER_CHANNEL_BUSY;

   if ((InputChannel>NB_CHANNELS) || (InputChannel<=0))
     return MODORDER_INVALID_CHANNEL;

   AddedLink=LinkList;
   while (AddedLink!=NULL){
     if ((AddedLink->Plug.ModuleIndex==ModuleIndex) && (AddedLink->Plug.InputChannel==InputChannel))
       return MODORDER_LINK_EXISTS;
     AddedLink=(TLinkList*) AddedLink->Next;
   };

   AddedLink=new TLinkList;
   AddedLink->Plug.Module=Module;
   AddedLink->Plug.ModuleIndex=ModuleIndex;
   AddedLink->Plug.InputChannel=InputChannel;
   AddedLink->Plug.State=State;

   // insert the link in the list SKIPPING the feedback
   TLinkList *l1, *l2;

   l2 = NULL;
   l1 = LinkList;
   while ((l1!=NULL) && (l1->Plug.State == LINK_STATE_FEEDBACK)){
     l2=l1;
     l1=(TLinkList*) l1->Next;
   };

   AddedLink->Next=(void*) l1;
   if (l2==NULL) 
     LinkList=AddedLink;
   else 
     l2->Next=(void*) AddedLink;

   ((CModule*) Module)->SetUsed(InputChannel);

   NbLinks++;
   return MODORDER_OK;
 }

 // Check if the connection list contains a connection to the given module
 // The module is given by its index.
 // This returns MODORDER_OK       if it is the case
 //              MODORDER_NO_LINK  if no such link is found
 int CLinkList::IsPluggedIn(int ModuleIndex){
   TLinkList *TmpList;
   TmpList=LinkList;
   while (TmpList!=NULL){
     if (TmpList->Plug.ModuleIndex==ModuleIndex)
       return MODORDER_OK;
    TmpList=(TLinkList*) TmpList->Next;
   };
   return MODORDER_NO_LINK;
 }

 // place the position pointer to the beginning
 // returns MODORDER_OK
 int CLinkList::Rewind(){
   Current=LinkList;

   return MODORDER_OK;
 }

 // return the current position link and increment the position
 TLinkTo* CLinkList::GetCurrentLink(){
   TLinkTo *Tmp;

   if (Current!=NULL){
     Tmp=&(Current->Plug);
     Current=(TLinkList*) Current->Next;
   } else 
     Tmp=NULL;

   return Tmp;
 }

 //---------------------------------------------------------------------------------
 // CModule class
 // This is the class dealing with the properties of a module
 //---------------------------------------------------------------------------------

 // ctor & dtor
 CModule::CModule(){
   Prop=NULL;
   IsGloballyConnected=0;
   Depth=0;
   UseTable=0;
 }

 CModule::~CModule(){
   delete[] Name;

   TProperty *tmp;
   while (Prop!=NULL){
     tmp=Prop->next;
     delete Prop;
     Prop=tmp;
   };
 }

 // Sets the module parameters
 // parameters are the description string or the file module 
 // location and its index
 // Names MODORDER_INPUT_MODULE and MODORDER_OUTPUT_MODULE describe
 // the input and output modules
 int CModule::SetParams(char *i_Name, char *i_Label, int i_Index){
   Name= new char[strlen(i_Name)+1];
   strcpy(Name, i_Name);
   Label= new char[strlen(i_Label)+1];
   strcpy(Label, i_Label);
   Index=i_Index;

   return MODORDER_OK;
 }

 // Add a connection to another module
 // The link is given by the output channel, a pointer to the destination 
 // module (as a void*) and its input channel
 // It first call TestLoop to take into account that the link may be a 
 // feedback link. In that case, no update of the structure is required.
 // It calls UpdateStruct for the output module to update the structure's depths
 // This returns MODORDER_OK              if successfull
 //              MODORDER_LINK_EXISTS     if link already exists
 //              MODORDER_INVALID_MODULE  if module index is not valid
 //              MODORDER_INVALID_CHANNEL if channel is not valid
 //              MODORDER_CHANNEL_BUSY    if channel is already used
 //              MODORDER_LOOP            if a loop has been found
 int CModule::AddLink(int OutputChannel, CModule *ToModule, int RemoteInputChannel, int State){
   int LinkReturn;

   // check if output channel is valid
   if ((OutputChannel<=0) || (OutputChannel>NB_CHANNELS))
     return MODORDER_INVALID_CHANNEL;

   // if the link is feedback, only process it if both ends are connected
   // to a module already connected to the structure
   if ((State == LINK_STATE_FEEDBACK) &&
       ((!IsGloballyConnected) || (!(ToModule->IsGloballyConnected))))
     return MODORDER_LOOP;

   // add link to the ad hoc link list.
   if ((LinkReturn=Links[OutputChannel-1].Append(ToModule->Index, ToModule, 
						 RemoteInputChannel, State))!=MODORDER_OK)
     return LinkReturn;

   // check loops unless the link is explicitely stated as feedback
   if (State != LINK_STATE_FEEDBACK){
     if (ToModule->UpdateStruct(Depth+1, Index)==MODORDER_LOOP)
       return MODORDER_LOOP_ERROR;
   }

   // This module and the destination module is now surely connected to the structure!
   IsGloballyConnected=1;
   ToModule->IsGloballyConnected=1;

   return MODORDER_OK;
 }

 // update the depth of the current module
 // If the NewDepth is <= current Depth, do nothing and return MODORDER_OK.
 // If the NewDepth is > current Depth, set Depth to NewDepth transmit to all
 // "child" connections" (with NewDepth+1 as argument).
 // The second parameter is used for loop detection. It contains the index of the
 // module where the update starts. If the chain fall back to this module,
 // we have looped.
 // If the current module index is ParentModuleIndex, return MODORDER_LOOP.
 // Else, transmit ParentModuleIndex to all outputs. Returns MODORDER_OK
 // if all of them returns MODORDER_OK, and MODORDER_LOOP if at least one of 
 // them returns MODORDER_LOOP. If all outputs are NULL, return MODORDER_OK.
 // This returns MODORDER_OK     if successfull
 //              MODORDER_LOOP   if a loop has been found
 int CModule::UpdateStruct(int NewDepth, int ParentModuleIndex){
   int NextStepReturn;
   int Channel;
   TLinkTo *Plug;
   CModule *PluggedIn;

   // update depth
   if (Depth<NewDepth){
     Depth=NewDepth;
   }

   // follow links
   NextStepReturn=MODORDER_OK;
   for (Channel=1;(Channel<=NB_CHANNELS) && (NextStepReturn==MODORDER_OK);Channel++){
     Links[Channel-1].Rewind();
     while ((NextStepReturn==MODORDER_OK) 
	    && ((Plug=Links[Channel-1].GetCurrentLink())!=NULL)){
       PluggedIn=(CModule*) Plug->Module;
       // prevent from following links already associated with feedback.
       if (Plug->State!=LINK_STATE_FEEDBACK){
	 if (PluggedIn->Index==ParentModuleIndex)
	   NextStepReturn=MODORDER_LOOP;
	 else
	   NextStepReturn=PluggedIn->UpdateStruct(NewDepth+1, ParentModuleIndex);
       } else {
	 // we are on a link already registered as feedback. Skip it !
	 NextStepReturn=MODORDER_OK;
       }
     };
   }

   return NextStepReturn;
 }

 // Check if the input channel passed as parameter is used.
 // returns MODORDER_USED or MODORDER_UNUSED.
 int CModule::IsUsed(int InputChannel){
   if (UseTable & (1<<(InputChannel-1)))
     return MODORDER_USED;

   return MODORDER_UNUSED;
 }

 // Sets the specified channel to used status.
 // Returns MODORDER_OK;
 int CModule::SetUsed(int InputChannel){
   UseTable|=(1<<(InputChannel-1));

   return MODORDER_OK;
 }

 // Add a property (parameter value for section 2)
 // Parameters are index and value.
 int CModule::AddProperty(int iIndex, char *iValue){
   TProperty *tmp;

   tmp=new TProperty;
   tmp->Index=iIndex;
   strcpy(tmp->Value, iValue);
   tmp->next=Prop;
   Prop=tmp;

   return MODORDER_OK;
 }



 /* Fills the table of modules and order them
 ** Given a file containing the list of modules, we fill the Modules 
 ** array with the list of modules ORDERED as they should linearily
 ** appear. The Modules array is supposed to be an allocated vector
 ** of pointers to CModules objects. A first call to the function
 ** passing NULL through Modules should only returns the nimber of
 ** modules through NbModules.
 ** This function returns MODORDER_OK              if successfull
 **                       MODORDER_FILE_NOT_EXISTS if file doesn't exists
 **                       MODORDER_READ_ERROR      if a syntax error occurs 
 **                       MODORDER_ERRORS          if some structure error is detected
 ** Outputs are only filled when successful ! 
 */
 int CreateStructure(char* FileName, CModule **Modules, 
		     int *NbModules, TModuleInfo *pInfo){

   FILE *flux;                 // stream associated with FileName
   int NbMod, NbLink;          // Number of modules and links
   int tmp, i, j, k;           // temporary variable and counters
   char Name[255];             // Module name read in the file
   char Label[255];            // Module label read in the file
   char PValue[20];            // string-like value for extra module property
   int **LinksTable;           // List between modules
   int NbAdd, HaveAdd;         // Control variables for link addition to structure
   CModule *TmpMod;            // Pointer to a CModule object for temporary usage during sorting
   char Tmp[255], State[255];  // link state
   int Continue;
   char *rc;

   // open file
   flux=fopen(FileName, "r");
   if (flux==NULL){
     ERROR1("File %s doesn't exists. Giving Up.\n", FileName);
     return MODORDER_FILE_NOT_EXISTS;
   }

   // check if the axp file contains descriptions 
   // i.e. the following informations:
   //   <project name>
   //   <description>
   //   <Nb In>
   //   <Nb Out>
   //   <input description>
   //   <output description>
   rc = fgets(pInfo->Name, 255, flux);
   if (rc==NULL){
     ERROR("Cannot read project name from script file. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }
   rc = fgets(pInfo->Descr, 255, flux);
   if (rc==NULL){
     ERROR("Cannot read project description from script file. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }
   tmp = fscanf(flux, "%d\n", &(pInfo->NbIn));
   if ((tmp==0) || (tmp==EOF)){
     ERROR("Cannot read number of input from script file. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }
   tmp = fscanf(flux, "%d\n", &(pInfo->NbOut));
   if ((tmp==0) || (tmp==EOF)){
     ERROR("Cannot read number of output from script file. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }
   rc = fgets(pInfo->InDescr, 255, flux);
   if (rc==NULL){
     ERROR("Cannot read inout description from script file. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }
   rc = fgets(pInfo->OutDescr, 255, flux);
   if (rc==NULL){
     ERROR("Cannot read outout description from script file. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }

   // removes linefeeds
   tmp = strlen(pInfo->Name)-1;
   while ((pInfo->Name[tmp]==0x0A) || (pInfo->Name[tmp]==0x0D)){ pInfo->Name[tmp]='\0'; tmp--;};

   tmp = strlen(pInfo->Descr)-1;
   while ((pInfo->Descr[tmp]==0x0A) || (pInfo->Descr[tmp]==0x0D)){ pInfo->Descr[tmp]='\0'; tmp--;};

   tmp = strlen(pInfo->InDescr)-1;
   while ((pInfo->InDescr[tmp]==0x0A) || (pInfo->InDescr[tmp]==0x0D)){ pInfo->InDescr[tmp]='\0'; tmp--;};

   tmp = strlen(pInfo->OutDescr)-1;
   while ((pInfo->OutDescr[tmp]==0x0A) || (pInfo->OutDescr[tmp]==0x0D)){ pInfo->OutDescr[tmp]='\0'; tmp--;};

   // retreive the number of modules
   tmp=fscanf(flux, "%d", &NbMod);
   if ((tmp==0) || (tmp==EOF)){
     ERROR("Cannot get the number of modules. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }
   if (NbMod<3){
     ERROR("We need at least 3 modules. Giving Up.\n");
     fclose(flux);
     return MODORDER_ERRORS;
   }

   (*NbModules)=NbMod;

   // ends here if only the number of modules is required
   if (Modules==NULL){
     fclose(flux);
     return MODORDER_OK;
   }

   OUTPUT1("%d modules\n", NbMod);

   // read all modules
   for (i=0;i<NbMod;i++){
     tmp=fscanf(flux, "%d. %s", &j, Name);
     if ((tmp==0) || (tmp==EOF)){
       ERROR1("Cannot read module number %d . Giving Up.\n", i+1);
       fclose(flux);
       return MODORDER_READ_ERROR;
     }
     if (j!=i+1){
       ERROR1("Error in link indexes (%d). Giving Up.\n", j);
       fclose(flux);
       return MODORDER_ERRORS;
     }
     // create the module with its parameters
     if (fscanf(flux, " | %s", Label)==0){
       int i,j,l;
       i=0;
       l=strlen(Name);
       while ((i<l) && (Name[i]!='/')) i++;
       if (i==l)
	 strcpy(Label,Name);
       else{
	 j=i+1;
	 i=j;
	 while ((i<l) && (Name[i]!='.')) i++;
	 l = j-i;
	 strncpy(Label, Name+j, l);
       }
     }
     fscanf(flux, " (%d-%d)", &j, &k);
     Modules[i]->SetParams(Name, Label, i+1);
     VERBOSE2("Module: '%s (%s)'\n", Modules[i]->Label, 
	      Modules[i]->Name); 
     // if module name is "IN", set it as connected
     if (strcmp(Name, "IN")==0)
       Modules[i]->IsGloballyConnected=1;

   }

   // retreive the number of links
   tmp=fscanf(flux, "%d", &NbLink);
   if ((tmp==0) || (tmp==EOF)){
     ERROR("Cannot get the number of links. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }
   OUTPUT1("%d links to read.\n", NbLink);
   if (fgets(Tmp, 255, flux)==NULL){
     ERROR("Cannot read links. Giving Up.\n");
     fclose(flux);
     return MODORDER_READ_ERROR;
   }

   // allocate memory for links table
   LinksTable=new int *[NbLink];

   // read links
   for (i=0;i<NbLink;i++){
     LinksTable[i]=new int[6]; // ModFrom(ChFrom) ->  ModTo(ChTo) + State + Added
     if (fgets(Tmp, 255, flux)==NULL){
       ERROR1("Cannot read link number %d. Giving Up.\n", i+1);
       for (j=0;j<i;j++) delete LinksTable[j];
       delete[] LinksTable;
       fclose(flux);
       return MODORDER_READ_ERROR;
     }

     tmp=sscanf(Tmp, "%d. %d(%d) -> %d(%d)", &j, 
		&(LinksTable[i][0]), &(LinksTable[i][1]), 
		&(LinksTable[i][2]), &(LinksTable[i][3]));

     // check that the reading has been done correctly
     if ((tmp==0) || (tmp==EOF)){
       ERROR1("Syntax error in link number %d. Giving Up.\n", i+1);
       for (j=0;j<i;j++) delete LinksTable[j];
       delete[] LinksTable;
       fclose(flux);
       return MODORDER_READ_ERROR;
     }

     // check the index
     if (j!=i+1){
       ERROR2("Error in link indexes (%d instead of %d). Giving Up.\n", j, i+1);
       for (j=0;j<i;j++) delete LinksTable[j];
       delete[] LinksTable;
       fclose(flux);
       return MODORDER_ERRORS;
     }

     // check if additional state information is given
     memset(State, 0, 255*sizeof(char));
     tmp=sscanf(Tmp, "%d. %d(%d) -> %d(%d) %s", &j, 
		&(LinksTable[i][0]), &(LinksTable[i][1]), 
		&(LinksTable[i][2]), &(LinksTable[i][3]),
		State);
     j = strlen(State);
     while ((State[j-1]==0x0A) || (State[j-1]==0x0D)){
       j--;
       State[j]=0;
     }
     LinksTable[i][4]=LINK_STATE_NORMAL;
     if (j>0){
       for (tmp=0;tmp<j;tmp++) State[tmp] = toupper(State[tmp]);
       if (strcmp(State, "FEEDBACK")==0){
	 VERBOSE1("Link %d stated as feedback.\n", i+1);
	 LinksTable[i][4]=LINK_STATE_FEEDBACK;
       }else{
	 ERROR2("Error in link %d: unknown state given ('%s').\n", i+1, State);
	return MODORDER_ERRORS;
      }
    }

    // everything seems ready at this stage
    VERBOSE4("Link from module %2d channel %2d to %2d channel %2d\n",
	     LinksTable[i][0], LinksTable[i][1], 
	     LinksTable[i][2], LinksTable[i][3]);
    LinksTable[i][5]=0;
  }

  // read the modules extra-properties
  Continue=1;
  while ((Continue) && (!feof(flux))){
    tmp=fscanf(flux, "%d %d %s", &i, &j, PValue);
    if ((tmp==0) || (tmp==EOF)){
      Continue=0;
    } else {
      if (i>NbMod){
	ERROR("Error in properties syntax. Giving Up.\n");
	for (j=0;j<i;j++) delete LinksTable[j];
	delete[] LinksTable;
	fclose(flux);
	return MODORDER_ERRORS;
      }
      Modules[i-1]->AddProperty(j,PValue);
    }
  };

  fclose(flux);

  // start adding modules one by one.
  NbAdd=0;
  do{
    HaveAdd=0;
    // browse all links
    for (i=0;i<NbLink;i++){
      // nothing to do if link already added
      if (LinksTable[i][5]==1)
	continue;
      // check if link start with a connected module
      if ((Modules[LinksTable[i][0]-1]->IsGloballyConnected==0) 
	  && (Modules[LinksTable[i][2]-1]->IsGloballyConnected==0))
	continue;
      // We thus have a link to add !
      switch (Modules[LinksTable[i][0]-1]->AddLink(LinksTable[i][1], 
						   Modules[LinksTable[i][2]-1],
						   LinksTable[i][3],
						   LinksTable[i][4])){
      case MODORDER_OK:              // this is a valid link
	VERY_VERBOSE1("Inserting link number %d\n", i+1);
	NbAdd+=HaveAdd=1;
	LinksTable[i][5]=1;
	break; 
      case MODORDER_LINK_EXISTS:     // link already exists	
	WARNING4("Warning: Link %d(%d) -> %d(%d) already defined\n",
		 LinksTable[i][0], LinksTable[i][1], 
		 LinksTable[i][2], LinksTable[i][3]);
	NbAdd+=HaveAdd=1;
	LinksTable[i][5]=1;
	break;
      case MODORDER_INVALID_CHANNEL: // channel is not valid
	ERROR4("Link %d(%d) -> %d(%d) : one of the channels is invalid.\n",
		LinksTable[i][0], LinksTable[i][1], 
		LinksTable[i][2], LinksTable[i][3]);
	for (j=0;j<NbLink;j++) delete LinksTable[j];
	delete[] LinksTable;
	return MODORDER_ERRORS;
	break;
      case MODORDER_CHANNEL_BUSY:    // channel is already used
	ERROR4("Link %d(%d) -> %d(%d) : input channel already used.\n",
		LinksTable[i][0], LinksTable[i][1], 
		LinksTable[i][2], LinksTable[i][3]);
	for (j=0;j<NbLink;j++) delete LinksTable[j];
	delete[] LinksTable;
	return MODORDER_ERRORS;
	break;
      case MODORDER_LOOP:            // not-yet-insertable feedback link
	break;
      case MODORDER_LOOP_ERROR:      // structure contains a loop !
	ERROR4("Fatal error: loop found when adding link %d(%d) -> %d(%d). Giving Up.\n",
	       LinksTable[i][0], LinksTable[i][1], 
	       LinksTable[i][2], LinksTable[i][3]);
	WARNING("You probably need to specify a link as feedback.\n");
	for (j=0;j<NbLink;j++) delete LinksTable[j];
	delete[] LinksTable;
	return MODORDER_ERRORS;
	break;
	/* still need to check that one !
	ERROR2("Multiple feedback from the same output not allowed (module %d, output %d)\n", LinksTable[i][0], LinksTable[i][1]);
	*/
      default:
	WARNING("Unknow return in link insertion\n");
      };
    }
  } while (HaveAdd!=0);

  if (NbAdd!=NbLink){
    ERROR2("Not all the links have been added (%d/%d).\n", NbAdd, NbLink);
    for (j=0;j<NbLink;j++) delete LinksTable[j];
    delete[] LinksTable;
    return MODORDER_ERRORS;
  } 

  // sorting links
  tmp=1;
  j=NbMod-1;
  while ((tmp==1) && (j>=0)){
    tmp=0;
    for (i=0;i<j;i++){
      if (Modules[i]->Depth>Modules[i+1]->Depth){
	TmpMod=Modules[i];
	Modules[i]=Modules[i+1];
	Modules[i+1]=TmpMod;
	tmp=1;
      }
    }
    j--;
  };

  for (j=0;j<NbLink;j++) delete LinksTable[j];
  delete[] LinksTable;

  return MODORDER_OK;
}
