//--------------------------------------------------------------------//
//  AgALag : Axoris group AL3101 application generator                //
//                                                                    //
//                 _         _    _      _                            //
//                / \____   / \  | |    / \____                       //
//               / _ \  _\ / _ \ | |   / _ \  _\                      //
//              / /_\ \  __ /_\ \| |  / /_\ \  __                     //
//             / _____ \ \ \____ \ |_/______ \ \ \                    //
//            /_/     \_\___\   \_\____|    \_\___\                   //
//                                                                    //
//  File: module.cpp                                                  //
//  Description: functions dealing with one particular module code    //
//  Version: 1.0                                                      //
//  Author: Gregor                                                    //
//  e-mail: Gregor.Soyez@swing.be                                     //
//                                                                    //
//  Copyright (C) 2002 Axoris                                         //
//--------------------------------------------------------------------//

#include "ProjAL.h"
#include "module.h"
#include "iomapping.h"
#include <string.h>
#include "tokenize.h"
#include "messages.h"
#include "extract.h"
#include <ctype.h>

/////////////////////////////////////////////////////////////////
// CCodeModule class implementation
/////////////////////////////////////////////////////////////////

// ctor & dtor
CCodeModule::CCodeModule(){
  int i;
  for (i=0;i<2*NB_CHANNELS;i++) CatchedTmpIO[i]=-1;
}

CCodeModule::~CCodeModule(){

}

/* 
** browse the source code and search for sections headers
** This fills the SectionsOffset table with sarting points 
** of every section in the file
** It returns MODULE_OK on success, MODULE_FILE_ERROR if 
** an error occurs with file and MODULE_MISSING_SECTION
** if some section seems to miss.
*/
int CCodeModule::CheckSections(){
  FILE *flux;
  int i, ok, end;
  long int position;
  char StrToFind[20], Line[255];

  // read module header
  VERBOSE1("Reading informations for module %s.\n", Name);
  if (GetModuleInfo(Name, &info)!=MODULE_HEADER_OK){
    ERROR1(" Module %s has a wrong header format.\n", Name);
    return MODULE_FILE_ERROR;
  }

  // start sections searching

  flux=fopen(Name, "r");
  if (flux==NULL){
    ERROR1("  Source file %s cannot be opened. Giving Up.\n", Name);
    return MODULE_FILE_ERROR;
  }
  
  position=0;

  //for (i=-3;i<NB_SECTIONS;i++){
  for (i=0;i<NB_SECTIONS;i++){
    ok=0;
    sprintf(StrToFind, ";; SEC%d", i+1);

    while (ok==0){
      if (fgets(Line, 255, flux)==NULL){
	// we reached EOF before having found all sections
	ERROR1("  End of source %s file reached without all sections !!\n", Name);
	return MODULE_MISSING_SECTION;
      }
      if (strstr(Line, StrToFind)!=NULL){
	ok=1;
	SectionsOffset[i]=position;
	VERY_VERBOSE2("  Section %d found at position %ld\n", i+1, position);
      }
      position=ftell(flux);
    };
  }

  fseek(flux, 0L, SEEK_END);
  SectionsOffset[NB_SECTIONS]=ftell(flux);

  fclose(flux);

  // create the definition string
  /*  ok=0;
  for (i=0;i<(int) strlen(Name);i++)
    if (Name[i]=='/') ok=i;
  sprintf(DefStr, "%c", Name[ok+1]);
  end = (int) strlen(Name);
  if ((Name[end-4]=='.') && (Name[end-3]=='a') && (Name[end-2]=='s') && (Name[end-1]=='m')) end-=4; 
  for (i=ok+2;i<end;i++){
    if (((Name[i]>='a') && (Name[i]<='z')) 
	|| ((Name[i]>='A') && (Name[i]<='Z')) 
	|| ((Name[i]>='0') && (Name[i]<='9')))
      strncat(DefStr, Name+i, 1);
  } 
  sprintf(Line, "%d", Index);
  strcat(DefStr, Line);*/
  sprintf(DefStr, "%s%d", Label, Index);

  return MODULE_OK;
}

/*
** Retreive the number of cycles used by the module 
** and the memory needed.
** These are given by 
** ; Cycles      : ? cycles (-1=unknownw)
** ; Memory      : ? words (needed by appl generator)
** in the module header.
*/
int CCodeModule::GetCycles(){
  return info.Cycles; //NbCycles;
}

int CCodeModule::GetMemoryNeeds(){
  return info.MemUsage; //MemNeeds;
}

int CCodeModule::GetDelayLine(){
  return info.UseDelay; //IsDelayLine;
}

int CCodeModule::SetMemOffset(int i_offset){
  MemoryOffset=i_offset;
  
  //  sprintf(MemDefStr, "%sMemOffset+", DefStr);
  return MODULE_OK;
}

int CCodeModule::GetMemOffset(){
  return MemoryOffset;
}

/*
** All the outputs which are not directly mapped on the OUTPUT module
** have to pass through a TMP? temporary variable. Moreover, all
** inputs coming from module different from line input have to
** be replaced by the TMP? output of the preceeding module.
** This function creates the table of these replacements . It
** uses the tools in iomapping.h. It returns MODULE_OK when 
** successful, MODULE_NO_TMP_LEFT if there are not enough free
** TMP? temporary variables and MODULE_ADD_ERROR if the link 
** for output could not be added. The case where an output is connected
** to another module and the output module at the same time
** is ambiguous ... MODULE_WRONG_STRUCTURE is returned.
** The outputs can be modified internally and the inputs are 
** modified by calling the second function for the next module
** from the first one. That second function return MODULE_ADD_ERROR 
** In case of troubles and MODULE_OK if everything's fine.
*/
int CCodeModule::CreateTemporaryIO(){ 
  int i, out;
  int OnlyOutputAllowed;
  char OutStr[10], ReplStr[10];
  TLinkTo *TmpLink, *FirstLink;
  bool IsFeedBack;

  // browse all outputs to know what they are connected to
  for (i=0;i<NB_CHANNELS;i++){
    OnlyOutputAllowed=0;
    Links[i].Rewind();
    FirstLink=TmpLink=Links[i].GetCurrentLink();

    // browse channel connections
    while (TmpLink!=NULL){
      // test if we connect to the output module
      if (strcmp(((CModule*) TmpLink->Module)->Name, OUTPUT_MODULE_NAME)==0){

	// deal with the special case of the input module
	if (strcmp(Name, INPUT_MODULE_NAME)==0){
	  ERROR("Cannot directly connect the input module to the output module.\n");
	  return MODULE_WRONG_STRUCTURE;
	}

	// check if we are already linked with an effect module
	if (TmpLink!=FirstLink){
	  if (OnlyOutputAllowed==0){
	    // we already have redirected to some effect module
	    ERROR("Cannot connect an output channel to both OUT module and effect module.\n");
	    FreeIO(out);
	    return MODULE_WRONG_STRUCTURE;
	  } else {
	    // we can only connect to one output channel
	    ERROR("Cannot connect an output channel to two different OUT module channels.\n");
	    //FreeIO(out);
	    return MODULE_WRONG_STRUCTURE;
	  }
	} else {
	  OnlyOutputAllowed=1;
	  // if the output channel is not the default one,
	  // add a replacement pattern
	  if ((i+1)!=TmpLink->InputChannel){
	    sprintf(OutStr, "OUT%d", (i+1));
	    sprintf(ReplStr, "OUT%d", TmpLink->InputChannel);

	    // Add a replacement for output  
	    if (Add(OutStr, ReplStr)!=REPLACE_OK)
	      return MODULE_ADD_ERROR;
	  }
	  TmpLink=Links[i].GetCurrentLink();
	}
      } else {
	// we are connected to an effect module

	// deal with the case of input module !
	if (strcmp(Name, INPUT_MODULE_NAME)==0){
	  // we need to check that the destination module reads the 
	  // correct input channel
	  if ((i+1)!=TmpLink->InputChannel){
	    sprintf(ReplStr, "IN%d", (i+1));
	    sprintf(OutStr, "IN%d", TmpLink->InputChannel);

	    // Add a replacement for output  
	    if (((CCodeModule*)TmpLink->Module)->Add(OutStr, ReplStr)!=REPLACE_OK)
	      return MODULE_ADD_ERROR;
	  }	  
	} else {
	  if (OnlyOutputAllowed!=0){
	    ERROR("Cannot connect an output channel to both OUT module and effect module.\n");
	    return MODULE_WRONG_STRUCTURE;
	  } else {
	    // check if the link has to be treated as feedback
	    if (TmpLink->State==LINK_STATE_FEEDBACK)
	      IsFeedBack = true;
	    else
	      IsFeedBack = false;

	    // set temporaryIO associated with the link.
	    if (TmpLink==FirstLink){
	      if (QueryIO(&out, IsFeedBack)!=IO_MAPPING_OK)
		return MODULE_NO_TMP_LEFT;
	      
	      sprintf(OutStr, "OUT%d", (i+1));
	      sprintf(ReplStr, "TMP%X", out);
	      
	      // Add a replacement for output  
	      if (Add(OutStr, ReplStr)!=REPLACE_OK)
		return MODULE_ADD_ERROR;
	    } else {
	      if (IsFeedBack){
		ERROR("Module with outgoing feedback has link-order problem\n");
		return MODULE_ADD_ERROR;
	      }
	    }
	    // and add replacements for inputs of all connected modules	    
	    ((CCodeModule*) TmpLink->Module)->AddTemporaryInput(TmpLink->InputChannel, out);  
	  }
	}
	// Jump to next link
	TmpLink=Links[i].GetCurrentLink();
      }
    };
  }
  
  return MODULE_OK;
}

int CCodeModule::AddTemporaryInput(int Input, int TempIOOffset){
  char In[20], Repl[20];

  sprintf(In, "IN%d", Input);
  sprintf(Repl, "TMP%X", TempIOOffset);
  CatchedTmpIO[Input-1]=TempIOOffset;
  if (Add(In, Repl)!=REPLACE_OK)
    return MODULE_ADD_ERROR;

  return MODULE_OK;
}

/*
** perform replacement of parameters in section 2
*/
int CCodeModule::UpdateProperty(int idx, char *Line){
  TProperty *tmp;
  int Continue=1;
  char ***tokens;
  int nb=0, i=0;

  tokens=new (char**);
  nb=StrTokenize(Line, tokens);
  if ((nb<3) || ((*tokens)[0][0]==';')
      || ((strcmp((*tokens)[1], "EQU")!=0) && (strcmp((*tokens)[1], "equ")!=0) && (strcmp((*tokens)[1], "Equ")!=0))){
    return MODULE_NOT_PROPERTY;
  }

  // search if this is present in property list
  tmp=Prop;
  while ((Continue) && (tmp!=NULL)){
    if (tmp->Index==idx){
      // got it ! perform replacement...
      sprintf(Line, "%s %s %s", (*tokens)[0], (*tokens)[1], tmp->Value);
      Continue=0;
    }
    tmp=tmp->next;
  };
  
  for (i=0;i<nb;i++) delete (*tokens)[i];
  delete[] (*tokens);
  delete tokens;
  return MODULE_OK;
}

/*
** Append the requested section to the given stream
** This function returns MODULE_OK on success
** and MODULE_SYNTAX_ERROR if a syntax error has been found.
** section is an integer between 1 and 7
*/
int CCodeModule::AppendSectionToStream(int section, FILE *OutputStream){
  FILE *fmod;
  char Line[255], NewLine[255], InitLine[20]; //, MemStr[160];
  int go=1;
  int idx, i;

  fmod=fopen(Name, "r");

  // update replacement list
  // sections 1 and 4 defines addresses which are affected by offset.
  // these sections are written in the Deal... function !
  switch (section){
  case 1: 
    if (DealCtrlParam(fmod, OutputStream)!=MODULE_OK) 
      return MODULE_SYNTAX_ERROR;
    fclose(fmod); 
    return MODULE_OK;
    break;
  case 2: if (DealInitVal  (fmod)!=MODULE_OK) return MODULE_SYNTAX_ERROR; break;
  case 3: if (DealCstDef   (fmod)!=MODULE_OK) return MODULE_SYNTAX_ERROR; break;
  case 4: 
    if (DealMemDef(fmod, OutputStream)!=MODULE_OK) 
      return MODULE_SYNTAX_ERROR; 
    fclose(fmod); 
    return MODULE_OK;
    break;
  case 5: if (DealInitMem  (fmod)!=MODULE_OK) return MODULE_SYNTAX_ERROR; break;
  case 6: if (DealInitSec  (fmod)!=MODULE_OK) return MODULE_SYNTAX_ERROR; break;
  case 7: if (DealCode     (fmod)!=MODULE_OK) return MODULE_SYNTAX_ERROR; break;
  };

  // return to the beginning of the section
  fseek(fmod, SectionsOffset[section-1], SEEK_SET);
  // the first line contains the section header !! pass it ...
  fgets(Line, 255, fmod);
  // for the 6th section (init) we have other line to read before starting to copy
  if (section==6){
    go=0;
    sprintf(InitLine, "EndInit");
  }

  idx=0;
  char ***tokens;
  int nb=0;
  char NewLine2[255];
  tokens=new (char**);
  bool isempty;
  
  while (ftell(fmod)<SectionsOffset[section]){
    fgets(Line, 255, fmod);
    if (go!=0){
      if ((section!=6) || (strstr(Line, InitLine)==NULL)){
	// perform replacement
	Perform(Line, NewLine);
	// check for property specification in section 2
	if (section==2){
	  if (UpdateProperty(idx+1, NewLine)==MODULE_OK) idx++;
	}
	// in section 6, remove multiple appearence of "C 0x1000000\nSCA 0x0 INIT"
	if (section==6){
	  nb=StrTokenize(NewLine, tokens);
	  if (!((nb>=2) && (CompareStrings((*tokens)[0], "C")==0))){
	    fprintf(OutputStream, "%s\n", NewLine);
	  } else {
	    if (!(ftell(fmod)<SectionsOffset[section]))
	      return MODULE_SYNTAX_ERROR;
	    // if next line is not INIT, we must write it
	    fgets(Line, 255, fmod);
	    Perform(Line, NewLine2);
	    for (i=0;i<nb;i++) delete (*tokens)[i];
	    delete[] (*tokens);
	    nb=StrTokenize(NewLine2, tokens);
	    if (!((nb>=3) && (CompareStrings((*tokens)[0], "SCA")==0) && (CompareStrings((*tokens)[2], "INIT")==0))){
	      // this is not the end of the section
	      fprintf(OutputStream, "%s\n%s\n", NewLine, NewLine2);
	    }
	  }
	  if (nb>0){
	    for (i=0;i<nb;i++) delete (*tokens)[i];
	    delete[] (*tokens);
	  }
	} else {
	  // in definition sections: do not write empty lines
	  if (section<6){
	    isempty=true;
	    for (i=0;i<strlen(NewLine);i++)
	      if ((NewLine[i]!=0x0A) && (NewLine[i]!=0x0D) && 
		  (NewLine[i]!=' ')  && (NewLine[i]!='\t')) 
		isempty = false;
	    if (!isempty)
	      fprintf(OutputStream, "%s\n", NewLine);
	  } else {
	    fprintf(OutputStream, "%s\n", NewLine);
	  }
	}
      }
    } else {
      // we are in init section and we're looking for 'EndInit'
      if (strstr(Line, InitLine)!=NULL){
	go=1;
	//sprintf(InitLine, "EndInit");
      }
    }
  };

  fclose(fmod);

  delete tokens;
  return MODULE_OK;
}

/* 
** Every Application & Module consists in 7 sections :
**   SEC1 Control parameters
**   SEC2 Initial values of control parameters (S3.24 format)
**   SEC3 Constants definitions
**   SEC4 Memory variables definition
**   SEC5 Initial values of memory variables
**   SEC6 Initialization
**   SEC7 Module code
** The following functions make the insertion work typical for 
** each section
*/
int CCodeModule::DealCtrlParam(FILE *flux, FILE *out){
  // all controls parameters are memory labels ...
  // Syntax: only empty lines, comments and ABS commands are allowed
  char Line[255];
  char Ptr[25];
  char ***tokens;
  int nb,i,len;
  UINT32 addr;

  tokens=new (char**);
  fseek(flux, SectionsOffset[0], SEEK_SET);
  while (ftell(flux)<SectionsOffset[1]){
    fgets(Line, 255, flux);
    nb=StrTokenize(Line, tokens);
    if ((nb>0) && ((*tokens)[0][0]!=';')){
      // lower case the 3rd token
      if (nb>=3){
	len = strlen((*tokens)[2]);
	for(i=0;i<len;i++) (*tokens)[2][i]=toupper((*tokens)[2][i]);
      }
      // check if it is an ABS statement
      if ((strcmp((*tokens)[0], "ABS")!=0) || (nb<3)
	  || (ExtractHexa((*tokens)[2], &addr, 0)!=0) ){
	//fprintf(stdout, "'%s'.\n", (*tokens)[2]);
	for (i=0;i<nb;i++) delete (*tokens)[i];
	delete[] (*tokens);
	return MODULE_SYNTAX_ERROR;
      } else {
	sprintf(Line, "%s%s", DefStr, (*tokens)[1]);
	Add((*tokens)[1], Line);
	fprintf(out, "ABS\t%s\t0x%X\n", Line, addr+MemoryOffset);
	// Add a replacement pattern @Line
	sprintf(Line, "@%s", (*tokens)[1]);
	sprintf(Ptr, "0x%X", addr+MemoryOffset);
	Add(Line,Ptr);
      }
      for (i=0;i<nb;i++) delete (*tokens)[i];
      delete[] (*tokens);
    }
  };

  delete tokens;
  return MODULE_OK;
}

int CCodeModule::DealInitVal(FILE *flux){
  // all values are cst definitions
  // Syntax: only empty lines, comments and EQU commands are allowed
  char Line[255];
  char ***tokens;
  int nb,i;

  tokens=new (char**);
  fseek(flux, SectionsOffset[1], SEEK_SET);
  while (ftell(flux)<SectionsOffset[2]){
    fgets(Line, 255, flux);
    nb=StrTokenize(Line, tokens);
    if ((nb>0) && ((*tokens)[0][0]!=';')){
      if ((nb<3) || (strcmp((*tokens)[1], "EQU")!=0)){
	for (i=0;i<nb;i++) delete (*tokens)[i];
	delete[] (*tokens);
	return MODULE_SYNTAX_ERROR;
      } else {
	// if the value name starts with an 'I', let it prefix DefStr.
	if (((*tokens)[0][0]=='I') || ((*tokens)[0][0]=='i'))
	  sprintf(Line, "I%s%s", DefStr, (*tokens)[0]+1);
	else
	  sprintf(Line, "%s%s", DefStr, (*tokens)[0]);
	Add((*tokens)[0], Line);
      }
      for (i=0;i<nb;i++) delete (*tokens)[i];
      delete[] (*tokens);
    }
  };

  delete tokens;
  return MODULE_OK;
}

int CCodeModule::DealCstDef(FILE *flux){
  // all values are cst definitions
  // Syntax: only empty lines, comments and EQU commands are allowed
  char Line[255];
  char ***tokens;
  int nb,i;

  tokens=new (char**);
  fseek(flux, SectionsOffset[2], SEEK_SET);
  while (ftell(flux)<SectionsOffset[3]){
    fgets(Line, 255, flux);
    nb=StrTokenize(Line, tokens);
    if ((nb>0) && ((*tokens)[0][0]!=';')){
      if ((nb<3) || (strcmp((*tokens)[1], "EQU")!=0)){
	for (i=0;i<nb;i++) delete (*tokens)[i];
	delete[] (*tokens);
	return MODULE_SYNTAX_ERROR;
      } else {
	sprintf(Line, "%s%s", DefStr, (*tokens)[0]);
	Add((*tokens)[0], Line);
      }
      for (i=0;i<nb;i++) delete (*tokens)[i];
      delete[] (*tokens);
    }
  };

  delete tokens;
  return MODULE_OK;
}

int CCodeModule::DealMemDef(FILE *flux, FILE *out){
  // all controls parameters are memory labels ...
  // Syntax: only empty lines, comments and ABS commands are allowed
  char Line[255];
  char Ptr[25];
  char ***tokens;
  int nb,i,len;
  UINT32 addr;

  tokens=new (char**);
  fseek(flux, SectionsOffset[3], SEEK_SET);
  while (ftell(flux)<SectionsOffset[4]){
    fgets(Line, 255, flux);
    nb=StrTokenize(Line, tokens);
    if ((nb>0) && ((*tokens)[0][0]!=';')){
      // lower case the 3rd token
      if (nb>=3){
	len = strlen((*tokens)[2]);
	for(i=0;i<len;i++) (*tokens)[2][i]=toupper((*tokens)[2][i]);
      }
      if ((strcmp((*tokens)[0], "ABS")!=0) || (nb<3)
	  || (ExtractHexa((*tokens)[2], &addr, 0)!=0) ){
	for (i=0;i<nb;i++) delete (*tokens)[i];
	delete[] (*tokens);
	return MODULE_SYNTAX_ERROR;
      } else {
	sprintf(Line, "%s%s", DefStr, (*tokens)[1]);
	Add((*tokens)[1], Line);
	fprintf(out, "ABS\t%s\t0x%X\n", Line, addr+MemoryOffset);
	// Add a replacement pattern @Line
	sprintf(Line, "@%s", (*tokens)[1]);
	sprintf(Ptr, "0x%X", addr+MemoryOffset);
	Add(Line,Ptr);
      }
      for (i=0;i<nb;i++) delete (*tokens)[i];
      delete[] (*tokens);
    }
  };

  delete tokens;
  return MODULE_OK;
}

int CCodeModule::DealInitMem(FILE *flux){
  // all values are cst definitions
  // Syntax: only empty lines, comments and EQU commands are allowed
  char Line[255];
  char ***tokens;
  int nb,i;

  tokens=new (char**);
  fseek(flux, SectionsOffset[4], SEEK_SET);
  while (ftell(flux)<SectionsOffset[5]){
    fgets(Line, 255, flux);
    nb=StrTokenize(Line, tokens);
    if ((nb>0) && ((*tokens)[0][0]!=';')){
      if ((nb<3) || (strcmp((*tokens)[1], "EQU")!=0)){
	for (i=0;i<nb;i++) delete (*tokens)[i];
	delete[] (*tokens);
	return MODULE_SYNTAX_ERROR;
      } else {
	// if the value name starts with an 'I', let it prefix DefStr.
	if (((*tokens)[0][0]=='I') || ((*tokens)[0][0]=='i'))
	  sprintf(Line, "I%s%s", DefStr, (*tokens)[0]+1);
	else
	  sprintf(Line, "%s%s", DefStr, (*tokens)[0]);
	Add((*tokens)[0], Line);
      }
      for (i=0;i<nb;i++) delete (*tokens)[i];
      delete[] (*tokens);
    }
  };

  delete tokens;
  return MODULE_OK;
}

int CCodeModule::DealInitSec(FILE *flux){
  // the only new replacements to perform are labels
  // No particular syntax rules
  char Line[255], Cpy[255];
  char ***tokens;
  int nb,i;

  tokens=new (char**);
  fseek(flux, SectionsOffset[5], SEEK_SET);
  while (ftell(flux)<SectionsOffset[6]){
    fgets(Line, 255, flux);
    nb=StrTokenize(Line, tokens);
    if ((nb>0) && ((*tokens)[0][strlen((*tokens)[0])-1]==':')){
      // label definition
      sprintf(Line, "%s%s", DefStr, (*tokens)[0]);
      Add((*tokens)[0], Line);
      strncpy(Cpy, Line, strlen(Line)-1);
      Cpy[strlen(Line)-1]='\0';
      strncpy(Line, (*tokens)[0], strlen((*tokens)[0])-1);
      Line[strlen((*tokens)[0])-1]='\0';
      Add(Line, Cpy);
      for (i=0;i<nb;i++) delete (*tokens)[i];
      delete[] (*tokens);
    }
  };

  delete tokens;
  return MODULE_OK;
}

int CCodeModule::DealCode(FILE *flux){
  // the only new replacements to perform are labels
  // No particular syntax rules
  char Line[255], Cpy[255];
  char ***tokens;
  int nb,i;

  tokens=new (char**);
  fseek(flux, SectionsOffset[6], SEEK_SET);
  while (ftell(flux)<SectionsOffset[7]){
    fgets(Line, 255, flux);
    nb=StrTokenize(Line, tokens);
    if ((nb>0) && ((*tokens)[0][strlen((*tokens)[0])-1]==':')){
      // label definition
      sprintf(Line, "%s%s", DefStr, (*tokens)[0]);
      Add((*tokens)[0], Line);
      strncpy(Cpy, Line, strlen(Line)-1);
      Cpy[strlen(Line)-1]='\0';
      strncpy(Line, (*tokens)[0], strlen((*tokens)[0])-1);
      Line[strlen((*tokens)[0])-1]='\0';
      Add(Line, Cpy);
      for (i=0;i<nb;i++) delete (*tokens)[i];
      delete[] (*tokens);
    }
  };

  delete tokens;
  return MODULE_OK;
}

int CCodeModule::InitRemoved(){
  FILE *fmod;
  char Line[255], InitLine[30];
  int go=0;
  int n=0;
  char ***tokens;
  int nb=0;
  int idx;

  fmod=fopen(Name, "r");

  // go to the beginning of the init section
  fseek(fmod, SectionsOffset[5], SEEK_SET);

  // the first line contains the section header !! pass it ...
  fgets(Line, 255, fmod);
  go=0;
  sprintf(InitLine, "EndInit");
  tokens=new (char**);
  
  while (ftell(fmod)<SectionsOffset[6]){
    fgets(Line, 255, fmod);
    if (go!=0){
      if (strstr(Line, InitLine)==NULL){
	nb=StrTokenize(Line, tokens);
	if ((nb>=3) && (CompareStrings((*tokens)[0], "SCA")==0) && (CompareStrings((*tokens)[2], "INIT")==0)){
	  n++;
	}
	if (nb>0){
	  for (idx=0;idx<nb;idx++) delete (*tokens)[idx];
	  delete[] (*tokens);
	}
      }
    } else {
      // we are in init section and we're looking for 'EndInit'
      if (strstr(Line, InitLine)!=NULL) go=1;
    }
  };

  fclose(fmod);

  delete tokens;
  return n;
}
