/* Assembler for AL3101 */

// general headers
#include <stdio.h>
#include <string.h>

// local headers
#include "../../common/1K_types.h"
#include "../base/Table.h"
#include "../base/PreAsm.h"
#include "../base/Options.h"
#include "../base/DbgFiles.h"

#define NB_MAX_COMMAND    1024  // maximal number of commands allowed by the chip.

#define MAX_TOKEN_SIZE     255  // Maximum allowed size of a token
#define MAX_CMDLINE_SIZE   255
#define NB_MAX_TOKENS        3  // mnemonic + 2 param

#define ASM_READ_ERROR      -1
#define ASM_PARSE_ERROR     -1  // parsing error

// Mnemonic Table defined in "AL3101Table.cpp"
extern TMnemonicTable MnemonicTable[];
TOption Options;

typedef struct{
  int LineIndex;     // reference of the line in the .asm file (only usefull for error reporting)
  char Command[MAX_TOKEN_SIZE];
  char *Params[NB_MAX_TOKENS-1];
  int NbParams;
} TPreAssembledCommand;

void PrintHelp()
{
  fprintf(stderr, "Axoris assembler for Alesis AL3101 DSP v1.0\n");
  fprintf(stderr, "Usage : asm <options> assembler_filename\n");
  fprintf(stderr, "  -h       : show this help message\n");
  fprintf(stderr, "  -o <file>: specify output filename\n");
  fprintf(stderr, "  -v       : verbose mode\n");
  fprintf(stderr, "Please send bug reports or comments to gregor@axoris.be\n\n");
}

int CheckFilename(char* fn){
  int len = strlen(fn);

  if (Options.verbose) fprintf(stdout, "Checking input filename validity\n");
  
  // check if '.asm' is present at end of fn. Add '.asm' if it is not the case
  if ((len<5) || (fn[len-1]!='m') || (fn[len-2]!='s') 
      || (fn[len-3]!='a') || (fn[len-4]!='.')){
    strcat(fn, ".asm");
    len+=4;
  }

  // check if the file exist
  FILE *flux=fopen(fn, "r");
  if (flux==NULL){
    fprintf(stderr, "File '%s' doesn't exist. Giving up.\n", fn);
    return 1;
  }
  fclose(flux);

  return 0;
}

// Parse command line into mnemonic and parameters or -1 if error
// returns the number of tokens found
int Parse(char *CmdLine, char *Token[]){
  int start, end;
  int n, i;

  // retreive the length of the command line
  n=strlen(CmdLine);

  // if CommandLine is an emtpy string, return 0
  if (n==0){
    return 0;
  }

  // if there is(are) EOL caracter(s), replace it(them) by null caracters
  while ((CmdLine[n-1]==0x0A) || (CmdLine[n-1]==0x0D)){
    CmdLine[n-1]='\0';
    n--;
    // if CommandLine is an emtpy string, return 0
    if (n==0)
      return 0;
  }
  
  start=0;
  i=0;

  do{   // we loop until we arrive at EOL or when we find a comment sign
    // find the beginning of the next token
    while ((CmdLine[start]=='\t') || (CmdLine[start]==' ')) start++;

    // if EOL or comment, return the number of tokens found
    if ((start==n) || (CmdLine[start]==';')) return i;

    // if there is no more toke nallowed, we face a typica lparse error
    if (i==NB_MAX_TOKENS) return ASM_PARSE_ERROR;

    // scan the beginning token ...
    end=start+1;
    while ((CmdLine[end]!='\t') && (CmdLine[end]!=' ') 
	   && (CmdLine[end]!=';') && (end<n)) end++;

    // ... and store it.
    strncpy(Token[i], CmdLine+start, end-start);
    Token[i][end-start]='\0';
    i++;

    // prepare for a next search.
    start=end; //+1;
  } while ((start<=n) && (i<=NB_MAX_TOKENS));

  return i;
}

/* 
** Get the opcode associated with mnemonic mn.
** return 0 if succeeded, 1 otherwise
** if the function succeed, the opcode is 
** returned through 'op'
** The index in MnemonicTable is returned through 'index'
*/
int GetOPCode(char *mn, UINT32 *op, int *index){
  int i=0;

  // browse the MnemonicTable commands
  do{
    if (strcmp(MnemonicTable[i].Mnemonic, mn)==0){
      (*index) = i;
      (*op) = MnemonicTable[i].OpCode;
      return 0;
    }      
    i++;
  } while (i<NB_OPCODES);

  return 1;
} 

/* 
** writes assembled result to 
** object file
*/
int AddToObj(UINT32 wCode, UINT32 lCode, FILE *flux_out){
  char OutChar=0;
  unsigned int t;
  int i;

  for (i=1;i>=0;i--){
    t= (wCode >> (4*i))%0x10;
    if (t<=9)
      OutChar='0'+(char) t;
    else
      OutChar='A'+(char) (t-10);
    fprintf(flux_out, "%c", OutChar);
  }
  for (i=7;i>=0;i--){
    t= (lCode >> (4*i))%0x10;
    if (t<=9)
      OutChar='0'+(char) t;
    else
      OutChar='A'+(char) (t-10);
    fprintf(flux_out, "%c", OutChar);
  }
  fprintf(flux_out, "\n");
  
  return 0;
}


// return -1 if an error occured while browsing source code,
// or the number of error encountered
int MainAssembleLoop(char *In_fn, char *Out_fn){
  FILE *flux_in  = fopen(In_fn,  "r");
  FILE *flux_out = fopen(Out_fn, "w+");
  
  int FileSize;                 // Size of the input file
  int ErrorNumber=0;            // Number of detected errors

  int LineIndex=0;               // index of the command line currently read
  int CmdIndex=0;                // index of the effective command 
  int NbCmd=0;                   // number of commands
  char CmdLine[255];             // command line in the input file
  char* Tokens[NB_MAX_TOKENS];   // Mnemonic+Params in the command line
  int nb_tok;                    // number of parameters + 1
  UINT32 opcode;                 // opcode associated with Mnemonic
  int opcode_index;              // index of the opcode in the Mnemonictable
  int flag;                      // extra parameter flag
  UINT32 wParam, lParam, Code;   // code for assembled parameters and instruction

  TPreAssembledCommand CCode[NB_MAX_COMMAND]; // store preassembled code

  if (Options.verbose) fprintf(stdout, "Assembling: initialization\n");
  
  sprintf(CmdLine, "");
  for (flag=0;flag<NB_MAX_TOKENS;flag++) Tokens[flag]=new char[MAX_TOKEN_SIZE];
    
  // calculating size of the input file
  fseek(flux_in, 0L, SEEK_END);
  FileSize=ftell(flux_in);
  fseek(flux_in, 0L, SEEK_SET);

  // Init pre-assembler
  InitPreAsm();

  /* 
  ** We perform assembling in two steps :
  ** In the first one, we read all lines from the file, for each of these, we tokenize it.
  ** We decide between 3 cases:
  **  1. comment or empty line (Nb_tokkens == 0)
  **  2. Preasm Command: this can be EQU, ABS or a label definition. We thus call the 
  **     preassembler during this first stage.
  **  3. Everything else should be a valid command and, unless the number of tokens is
  **     wrong (>3), we store command in the CCode array !
  ** The second step consist in assembling the CCode informationg through the "pure" assembler.
   */

  // STEP 1.

  if (Options.verbose) fprintf(stdout, "Assembling: parsing file and pre-assembling\n");
  
  // browse command lines until EOF reached or an error occured
  while ((NbCmd<1024) && (fgets(CmdLine, MAX_CMDLINE_SIZE, flux_in)!=NULL)){

    LineIndex++;

    // convert all upper case ...
    for (nb_tok=0;nb_tok<(int) strlen(CmdLine);nb_tok++){
      if ((CmdLine[nb_tok]>='a') && (CmdLine[nb_tok]<='z')){
	CmdLine[nb_tok]-=('a'-'A');
      }
    }

    // parse command line into tokens
    if ((nb_tok=Parse(CmdLine, Tokens))==ASM_PARSE_ERROR){
      ErrorNumber++;
      fprintf(stderr,"%d: Parse error : too much paramerters\n", LineIndex);
      continue;
    }
    if (nb_tok==0) continue;

    // Pre-assemble string.
    switch(PreAssemble(nb_tok, Tokens, NbCmd)){
    case PREASM_NOTHING: // not a comment nor a preasm command -> store for step 2
      CCode[NbCmd].LineIndex=LineIndex;
      strcpy(CCode[NbCmd].Command, Tokens[0]);
      CCode[NbCmd].NbParams=nb_tok-1;
      CCode[NbCmd].Params[0]=new char[MAX_TOKEN_SIZE];
      CCode[NbCmd].Params[1]=new char[MAX_TOKEN_SIZE];
      for (nb_tok=0;nb_tok<CCode[NbCmd].NbParams;nb_tok++)
	strcpy(CCode[NbCmd].Params[nb_tok], Tokens[nb_tok+1]);
      NbCmd++;
      break;
    case PREASM_OK:      // preasm command treated by PreAssemble. Skip to next line.
      break;
    case PREASM_MAX_CST: // Preasm command ... preassembler capacity exceeded
      ErrorNumber++;
      fprintf(stderr,"%d: Preassembler reached maximal capacity\n", LineIndex);
      break;
    case PREASM_SYNTAX:  // preassembler command with syntax error
      ErrorNumber++;
      fprintf(stderr,"%d: Wrong preassembler syntax\n", LineIndex);
      break;
    case PREASM_FORMAT:  // preassembler command with format error
      ErrorNumber++;
      fprintf(stderr,"%d: Wrong preassembler format\n", LineIndex);
      break;
    case PREASM_EXISTS:  // already defined preassembler constant
      ErrorNumber++;
      fprintf(stderr,"%d: %s already defined\n", LineIndex, Tokens[0]);
      break;
    };

    sprintf(CmdLine, "");
  };

  // release memory for tokens
  for (nb_tok=0;nb_tok<NB_MAX_TOKENS;nb_tok++) delete[] Tokens[nb_tok];

  // if we have not read 'Filesize bytes, an error occured
  if (ftell(flux_in)<FileSize){
    fclose(flux_in);
    fprintf(stderr, "Error before EOF ... Giving up.\n\n");
    return ASM_READ_ERROR;
  }

  // close file
  fclose(flux_in);

  // STEP 2

  if (Options.verbose) fprintf(stdout, "Assembling: creating object code\n");
  
  for (CmdIndex=0;CmdIndex<NbCmd;CmdIndex++){

    // retreive opcode
    if (GetOPCode(CCode[CmdIndex].Command, &opcode, &opcode_index)!=0){
      ErrorNumber++;
      fprintf(stderr,"%d: Unrecognized mnemonic '%s'\n", 
	      CCode[CmdIndex].LineIndex, CCode[CmdIndex].Command);
      continue;
    }

    // call assembling function associated with opcode
    // this fills the parameter fields in the AL3101_disasm structure passed
    // as argument
    switch ((Code=MnemonicTable[opcode_index].f_asm(CCode[CmdIndex].NbParams, &(CCode[CmdIndex].Params[0]), &wParam, &lParam))){
    case -1:  // wrong number of parameter
      fprintf(stderr, "%d: '%s' doesn't take %d parameters\n", 
	      CCode[CmdIndex].LineIndex, CCode[CmdIndex].Command, CCode[CmdIndex].NbParams);
      ErrorNumber++;
      break;
    case 1:   // error in parameter 1
    case 2:   // error in parameter 2
      fprintf(stderr,"%d: Error in parameter %d\n", CCode[CmdIndex].LineIndex, Code);
      ErrorNumber++;
      break;
    default:   // call successful
      // complete the binary code with the opcode
      wParam += opcode<<2;
      // and fill the object file
      AddToObj(wParam, lParam, flux_out);
      break;
    }

    // Update PreAsm Program Counter
    PreAsmStepPC();
  }

  if (Options.verbose) fprintf(stdout, "Assembling: finishing\n");
  
  // free memory allocate for parameters inside the CCode array
  for (CmdIndex=0;CmdIndex<NbCmd;CmdIndex++){
    delete[] CCode[CmdIndex].Params[0];
    delete[] CCode[CmdIndex].Params[1];
  }

  // Close PreAsm !
  PreAsmClose();

  // show number of errors
  if (ErrorNumber==0){
    fprintf(stderr, "Assembling successfully completed (%d code lines)\n\n",
	    NbCmd);
  } else {
    fprintf(stderr, "Assembling found %d Errors\n\n", ErrorNumber);    
  }

  return ErrorNumber;
}

int main(int argc, char *argv[]){
  char optdescr;
  
  // retreive options:
  while ((optdescr = getopt(argc, argv, "ho:vm")) != -1){
    switch (optdescr){
    case 'h': PrintHelp(); return 0; break;
    case 'v': Options.verbose = 1; break;
    case 'm': Options.map = 1; break;
    case 'o': Options.output = 1; Options.outname = optarg; Options.n++; break;
    case ':': PrintHelp(); fprintf(stderr, "Missing argument. Giving up\n"); return 1; break;
    case '?': PrintHelp(); fprintf(stderr, "Unknown option. Giving up\n"); return 1; break;
    default: return 1; break;
    };
    Options.n++;
  }

  // check the number of arguments passed to '1kams'
  if (argc!=Options.n+2){
    PrintHelp();
    return 1;
  }

  // retreive the assembler filename
  int len = strlen(argv[argc-1]);
  char InputFilename[FILENAME_MAX];
  strcpy(InputFilename, argv[argc-1]);
  // check if it is a valid filename
  if (CheckFilename(InputFilename)!=0)
    return 1;

  // retreive the object filename or construct it.
  char *OutputFilename;
  if (Options.output){      // furnished by user
    OutputFilename = new char[strlen(Options.outname)+1];
    strcpy(OutputFilename, Options.outname);
  } else {           // construct it from assembler filename
    len = strlen(InputFilename);
    OutputFilename = new char[len+1];
    strcpy(OutputFilename, InputFilename);
    OutputFilename[len-3]='o';
    OutputFilename[len-2]='b';
    OutputFilename[len-1]='j';
  }

  // enter into the main disassemble loop
  int Result=MainAssembleLoop(InputFilename, OutputFilename);

  // generate map file
  if ((Result==0) && (Options.map)){
    char mapname[255];
    FILE *mapflux;
    strcpy(mapname, OutputFilename);
    len = strlen(OutputFilename);
    if ((OutputFilename[len-3]=='o') &&
        (OutputFilename[len-2]=='b') &&
        (OutputFilename[len-1]=='j'))
      len-=4;

    mapname[len  ]='.';
    mapname[len+1]='m';
    mapname[len+2]='a';
    mapname[len+3]='p';
    mapname[len+4]='\0';
    mapflux=fopen(mapname, "w+");
    GenerateMap(mapflux);
    fclose(mapflux);
  }

  // if assembling not successfull, remove object file
  if (Result!=0)
    remove(OutputFilename);

  // don't forget to release memory ...
  delete[] OutputFilename;

  return Result;
}
