/* Assembler for AL3101 */

#include <stdio.h>
#include <string.h>

#include "../../common/1K_types.h"
#include "../base/Table.h"
#include "../base/Options.h"

#define MAX_CMDLINE_SIZE   255
#define ASM_READ_ERROR      -1
#define NB_MAX_TOKENS        3  // mnemonic + 2 param

extern TMnemonicTable MnemonicTable[];
TOption Options;

char Comments[64][50]={
  "     C *[M]",            // CM
  "     C *[M], [M] -> B",  // LCM
  " A + C *[M]",            // CMA
  " A + C *[M], [M] -> B",  // LCMA
  " B + C *[M]",            // CMB
  "     C *[M],   A -> B",  // XCM
  " B + A *[M]",            // AMB
  " A + C *[M],   A -> B",  // XCMA
  " C + A *[M]",             // AMC
  " C + A *[M], [M] -> B",   // LAMC
  " C + B *[M]",             // BMC
  " C +[M]*[M], [M] -> B",   // LMMC
  " C +[M]*[M]",             // MMC
  " C +[M]*[M],   A -> B",   // XMMC
  " C +[M]",                 // 1MC
  " C +[M]    , [M] -> B",   // L1MC
  " C + A * A",              // AAC
  " C + A * A , [M] -> B",   // LAAC
  " C + B * A",              // BAC
  " B + C * A",              // CAB
  " D + C * A",              // CAD
  "     C * A , [M] -> B",   // LCA
  "[M]+ C * A",              // CAM
  " C + D * A",              // DAC
  " C + B * B",              // BBC
  "     C * B",              // CB
  "[M]+ C * B",              // CBM
  " C + A     ,   A -> B",   // X1AC
  " C + A",                  // 1AC
  " C + A     , [M] -> B",   // L1AC
  "LOG16(B)",                // LOGB
  "16^B",                    // EXPB
  "     C * A ,   A -> [M]", // SCA
  "     C * A ,   A -> [M], A -> B", // SXCA
  "     C * B ,   A -> [M]", // SCB
  " A + B * U ,   A -> [M]", // SBUA
  " A + C * U ,   A -> [M]", // SCUA
  " B + C * U ,   A -> [M]", // SCUB
  "     C * U ,   A -> [M]", // SCU
  " C + B * U ,   A -> [M]", // SBUC

  " C + A * A ,   A -> [M]", // SAAC
  " C + A * A ,   A -> [M], A -> B", // SXAAC
  " C + B * A ,   A -> [M]", // SBAC
  " C + B * B ,   A -> [M]", // SBBC
  " A + C * B ,   A -> [M]", // SCBA
  " B + C * A ,   A -> [M]", // SCAB
  "LOG16(B)   ,   A -> [M]", // SLOGB
  "16^B       ,   A -> [M]", // SEXPB
  " C + A * U ,   A -> [M]", // SAUC
  " C + A     ,   A -> [M]", // S1AC
  " C + A     ,   A -> [M], A -> B", // SX1AC
  " C + B     ,   A -> [M]", // S1BC
  " A + 0 * U ,   A -> [C&B + M]",   // SINDF
  " A + 0 * U ,   A -> [C&B + M]",   // SINDI
  " 0 -> U, B&A",   // ANDB
  " 0 -> U, C&A",   // ANDC
  "INTEGER : A + B",   // ADDB
  "INTEGER : A - B",   // SUBB
  "INTEGER : A * B",   // MLTB
  "if C, C -> A, else 0 -> A",   // C
  "if C, treat next C instruction as NOPs",   // SKIP
  " C + B",            // 1BC
  "[C&B + M]",   // LINDF
  "[C&B + M]"    // LINDI
};  

void PrintHelp()
{
  fprintf(stderr, "Axoris disassembler for Alesis AL3101 DSP v1.0\n\n");
  fprintf(stderr, "Usage : disasm <options> object_filename\n");
  fprintf(stderr, "  -d       : use decimal format for constants instead of hexadecimal\n");
  fprintf(stderr, "  -h       : show this help message\n");
  fprintf(stderr, "  -o <file>: specify output filename\n");
  fprintf(stderr, "  -v       : verbose mode\n");
  fprintf(stderr, "Please send bug reports or comments to gregor.soyez@swing.be\n\n");
}

int CheckFilename(char* fn){
  int len = strlen(fn);
  
  if (Options.verbose) fprintf(stdout, "Checking input filename validity\n");
  
  // check if '.obj' is present at end of fn. Add '.obj' if it is not the case
  if ((len<5) || (fn[len-1]!='j') || (fn[len-2]!='b') 
      || (fn[len-3]!='o') || (fn[len-4]!='.')){
    strcat(fn, ".obj");
    len+=4;
  }

  // check if the file exist
  FILE *flux=fopen(fn, "r");
  if (flux==NULL){
    fprintf(stderr, "File '%s' doesn't exist. Giving up.\n", fn);
    return 1;
  }
  fclose(flux);

  return 0;
}

/* check if the object line is valid :
**   1. has 10 caracters
**   2. is only composed of hexadecimal caracters
** Return 0 if sucess, i if step i has failed.
*/
int CheckLine(char *CmdLine, int len){
  int i;

  // removing EOL caracter(s)
  while ((CmdLine[len-1]==0x0A) || (CmdLine[len-1]==0x0A)){
    CmdLine[len-1]='\0';
    len--;
  }
  if (len!=10)
    return 1;

  for (i=0;i<10;i++){
    if ((CmdLine[i]<'0')
	|| ((CmdLine[i]>'9') && (CmdLine[i]<'A'))
	|| ((CmdLine[i]>'F') && (CmdLine[i]<'a'))
	|| (CmdLine[i]>'f')){
      return 2;
    }
  }
  
  return 0;
}

/* convert the object line string
** in integer as wParam (bis 33 to 40)
** and lParam (bits 1 to 32)
*/
int ConvertToInt(char *Cmd, UINT32 *wparam, UINT32 *lparam){
  int i;
  UINT32 val=0;
  (*wparam)=(*lparam)=0;

  for (i=0;i<10;i++){
    switch (Cmd[i]){
    case '0': val = 0; break;    case '1': val = 1; break;
    case '2': val = 2; break;    case '3': val = 3; break;
    case '4': val = 4; break;    case '5': val = 5; break;
    case '6': val = 6; break;    case '7': val = 7; break;
    case '8': val = 8; break;    case '9': val = 9; break;
    case 'A': case 'a': val = 10; break;
    case 'B': case 'b': val = 11; break;
    case 'C': case 'c': val = 12; break;
    case 'D': case 'd': val = 13; break;
    case 'E': case 'e': val = 14; break;
    case 'F': case 'f': val = 15; break;
    }
    if (i>1)
      (*lparam)+=(val<<(4*(9-i)));
    else
      (*wparam)+=(val<<(4*(1-i)));
  }

  return 0;
}

/* return 0 if succeeded, 1 otherwise
** if the function succeed, the opcode is 
** returned through 'op'
*/
int GetOPCode(UINT32 wparam, int *op_index){
  int i=0;
  UINT32 op;

  // retreive OpCode
  op = (wparam>>2);

  // browse the MnemonicTable
  do{
    if (MnemonicTable[i].OpCode==op){
      (*op_index) = i;
      return 0;
    }      
    i++;
  } while (i<NB_OPCODES);

  return 1;
} 

/* 
** writes assembled result to 
** object file
*/
int AddToAsm(char *AsmLine, FILE *flux_out){
  fprintf(flux_out, "%s", AsmLine);

  return 0;
}


// return -1 if an error occured while browsing source code,
// or the number of error encountered
int MainAssembleLoop(char *In_fn, char *Out_fn){
  FILE *flux_in  = fopen(In_fn,  "r");
  FILE *flux_out = fopen(Out_fn, "w+");
  
  long int FileSize;     // Size of the input file
  int ErrorNumber=0;     // Number of detected errors

  char CmdLine[255];     // command line in the input file
  char AsmLine[255];     // command line in the input file
  char ParamLine[255];   // command line in the input file
  int LineIndex=0;       // index of the current line in the file 
  int len;               // length of the Command line
  int opcode_index;      // index of the opcode in the Mnemonictable
  UINT32 wParam, lParam; // code for assembled parameters and instruction
  
  int Code;

  sprintf(CmdLine, "");
  if (Options.dec)
    SetConstantFormat(CST_FORMAT_DECIMAL);
  else 
    SetConstantFormat(CST_FORMAT_HEXADECIMAL);
    
  // calculating size of the input file
  fseek(flux_in, 0L, SEEK_END);
  FileSize=ftell(flux_in);
  fseek(flux_in, 0L, SEEK_SET);

  if (Options.verbose) fprintf(stdout, "Disassembling: processing ...\n");
  
  // browse command lines until EOF reached or an error occured
  while ((LineIndex<1024) && (fgets(CmdLine, MAX_CMDLINE_SIZE, flux_in)!=NULL)){
    LineIndex++;
    
    // check the syntax of the line
    len = strlen(CmdLine);
    switch(CheckLine(CmdLine, len)){
    case 1:
      ErrorNumber++;
      fprintf(stderr, "%d: Line has not a correct length\n", LineIndex);
      continue;
      break;
    case 2:
      ErrorNumber++;
      fprintf(stderr, "%d: Line contains non-valid caracters\n", LineIndex);
      continue;
      break;      
    }
    
    ConvertToInt(CmdLine, &wParam, &lParam);

    // get Mnemonic and get the associated opcode
    if (GetOPCode(wParam, &opcode_index)!=0){
      ErrorNumber++;
      fprintf(stderr, "%d: Unrecognized opcode '%x'\n", LineIndex, (wParam>>2));
      continue;
    }
    wParam%=0x4;

    switch ((Code=MnemonicTable[opcode_index].f_disasm(wParam, lParam, ParamLine))){
    case -1:
    case -2:
      fprintf(stderr, "%d: Error in parameter %d\n", LineIndex, Code);
      ErrorNumber++;
      break;
    default:
      sprintf(AsmLine, "%s\t%s\t; %s\n",
	      MnemonicTable[opcode_index+Code].Mnemonic,
	      ParamLine,
	      Comments[opcode_index+Code]);
      AddToAsm(AsmLine, flux_out);
      break;
    }

    sprintf(CmdLine, "");
  };

  if (ftell(flux_in)<FileSize){
    fclose(flux_in);
    fclose(flux_out);
    fprintf(stderr, "Error before EOF\n\n");
    return ASM_READ_ERROR;
  }

  // close file, show number of errors
  fclose(flux_in);
  fclose(flux_out);
  if (ErrorNumber==0)
    fprintf(stderr, "Disassembling sucessfully completed\n\n");
  else
    fprintf(stderr, "Disassembling found %d Errors\n\n", ErrorNumber);    
    
  return ErrorNumber;
}

int main(int argc, char *argv[]){
  char optdescr;
  
  // retreive options:
  while ((optdescr = getopt(argc, argv, "dho:v")) != -1){
    switch (optdescr){
    case 'd': fprintf(stdout, "Warning: disassembling in decimal format may lead to rounding errors.\n"); Options.dec=1; break;
    case 'h': PrintHelp(); return 0; break;
    case 'v': Options.verbose = 1; break;
    case 'o': Options.output = 1; Options.outname = optarg; Options.n++; break;
    case ':': PrintHelp(); fprintf(stderr, "Missing argument. Giving up\n"); return 1; break;
    case '?': PrintHelp(); fprintf(stderr, "Unknown option. Giving up\n"); return 1; break;
    default: return 1; break;
    };
    Options.n++;
  }

  // check the number of arguments passed to '1kams'
  if (argc!=Options.n+2){
    PrintHelp();
    return 1;
  }

  // retreive the assembler filename
  int len = strlen(argv[argc-1]);
  char InputFilename[FILENAME_MAX];
  strcpy(InputFilename, argv[argc-1]);
  // check if it is a valid filename
  if (CheckFilename(InputFilename)!=0)
    return 1;

  // retreive the object filename or construct it.
  char *OutputFilename;
  if (Options.output){      // furnished by user
    OutputFilename = new char[strlen(Options.outname)+1];
    strcpy(OutputFilename, Options.outname);
  } else {           // construct it from assembler filename
    len = strlen(InputFilename);
    OutputFilename = new char[len+1];
    strcpy(OutputFilename, InputFilename);
    OutputFilename[len-3]='a';
    OutputFilename[len-2]='s';
    OutputFilename[len-1]='m';
  }

  // enter into the main disassemble loop
  int Result=MainAssembleLoop(InputFilename, OutputFilename);

  // if assembling not successfull, remove object file
  if (Result!=0)
    remove(OutputFilename);

  // don't forget to release memory ...
  delete[] OutputFilename;

  return Result;
}
