// FileIO_PCM.cpp: implementation of the PCM File functions.
//
//////////////////////////////////////////////////////////////////////

#include "FileIO_PCM.h"
#include "FileIO_PCM_internal.h"

#include <stdlib.h>		// for the malloc function
#include <string.h>		// for the string routines in error functions


//////////////////////////////////////////////////////////////////////
// init & clean functions
//////////////////////////////////////////////////////////////////////
INT32 FileIO_PCM_Init(void **pInstance, CHAR *FileName, INT32 Mode, FileIO_SndCfg *cfg, INT32 *LeftSampleMemory, INT32 *RightSampleMemory)
{
	FileIO_PCM_Struct *pInst;

	// checking and getting file characteristics
	if (cfg->NumberOfBytesInSample != 2) return FILEIO_TYPE_NOT_SUPPORTED;
	if ((cfg->NumberOfChannels != 1) && (cfg->NumberOfChannels != 2)) return FILEIO_TYPE_NOT_SUPPORTED;
	// moreover checking if The number of channel wanted is not OK with what available in the file
	if (((RightSampleMemory == NULL) && (cfg->NumberOfChannels == 2)) || ((RightSampleMemory != NULL) && (cfg->NumberOfChannels != 2))) return FILEIO_WRONG_CHANNEL_NUMBER;

	// Allocate the space for the instance structure
	pInst = new FileIO_PCM_Struct;
	if (pInst == NULL) return FILEIO_CANNOT_ALLOCATE_INSTANCE;
	*pInstance = (void *)pInst;

	pInst->Mode = Mode;

	// open the file
	pInst->NumberOfChannels = cfg->NumberOfChannels;		// Number of channels contained in the file
	pInst->SamplingFrequency = cfg->SamplingFrequency;		// The sampling frequency of the file
	if (pInst->Mode != 0)
	{
		pInst->file = fopen(FileName, "w+b");
	}
	else
	{
		pInst->file = fopen(FileName, "rb");
	}
	if (pInst->file == NULL)
	{
		delete pInst;
		return FILEIO_CANNOT_OPEN_FILE;
	}
	pInst->Loop = 0;					// By default, no looping


	// do the DSP memory related stuff initialization
	pInst->LeftData = LeftSampleMemory;			// Pointer to the left sample memory
	pInst->RightData = RightSampleMemory;		// Pointer to the right sample memory
	
	return FILEIO_NOERROR;	// no problem occurs
}

INT32 FileIO_PCM_Clean(void *pInstance)
{
	FileIO_PCM_Struct *pInst = (FileIO_PCM_Struct *)pInstance;

	// closing the file	
	if (pInst->file != NULL) if (fclose(pInst->file) != 0) return FILEIO_CANNOT_CLOSE_FILE;

	// freeing the instance structure
	if (pInst != NULL) delete pInst;

	return FILEIO_NOERROR;
}


//////////////////////////////////////////////////////////////////////
// "processing" functions
//////////////////////////////////////////////////////////////////////
INT32 FileIO_PCM_ReadSample(void *pInstance)
{
	FileIO_PCM_Struct *pInst = (FileIO_PCM_Struct *)pInstance;
	INT16 Buffer[2];
	INT32 tempint;

	// Check if we try to read in a file which has been opened in that mode
	if (pInst->Mode != 0)
	{
		return FILEIO_WRONG_FILE_MODE;			// Try to read in a file that is in write mode
	}
	
	
	// Check if data is still available in the buffer
	INT32 result = fread(Buffer, sizeof(INT16), pInst->NumberOfChannels, pInst->file);
	if (result < (INT32)pInst->NumberOfChannels)
	{
		if (!feof(pInst->file))
		{
			return FILEIO_READ_ERROR;			// A problem occurs in ReadFile
		}
		else
		{
			if (pInst->Loop) fseek(pInst->file, 0, SEEK_SET);
			else return FILEIO_EOF_REACHED;		// End of file reached, reset to beginning of file
		}
	}

	// let's now transfer data from the buffer to the memory
	tempint = Buffer[0];							// first copy the 16 bits value in a 32 bits one
	tempint = tempint << 8;							// then prepare the data to fit a S3.24 format
	*(pInst->LeftData) = tempint;					// copy left/mono channel to memory
	if (pInst->NumberOfChannels == 2)				// check if there's a second channel
	{
		tempint = Buffer[1];						// first copy the 16 bits value in a 32 bits one
		tempint = tempint << 8;						// then prepare the data to fit a S3.24 format
		*(pInst->RightData) = tempint;				// copy right channel to memory
	}

	
	return FILEIO_NOERROR;	// no problem occurs
}

	
INT32 FileIO_PCM_WriteSample(void *pInstance)
{
	FileIO_PCM_Struct *pInst = (FileIO_PCM_Struct *)pInstance;
	INT16 Buffer[2];
	INT32 tempint;

	// Check if we try to read in a file which has been opened in that mode
	if (pInst->Mode == 0)
	{
		return FILEIO_WRONG_FILE_MODE;				// Try to write in a file that is in read mode
	}

	// let's transfer data from the memory to the buffer
	tempint = *(pInst->LeftData);
	if (tempint > (INT32)0x007FFFFF) tempint = 0x007FFFFF;
	if (tempint < (INT32)0xFF800000) tempint = 0xFF800000;
	tempint = tempint >> 8;
	Buffer[0] = (INT16)tempint;							// copy left/mono channel to buffer
	if (pInst->NumberOfChannels == 2)					// check if there's a second channel
	{
		tempint = *(pInst->RightData);
		if (tempint > (INT32)0x007FFFFF) tempint = 0x007FFFFF;
		if (tempint < (INT32)0xFF800000) tempint = 0xFF800000;
		tempint = tempint >> 8;
		Buffer[1] = (INT16)tempint;						// copy right channel to buffer
	}

	INT32 result = fwrite(Buffer, sizeof(INT16), pInst->NumberOfChannels, pInst->file);
	if (result != (INT32)pInst->NumberOfChannels)
	{
		return FILEIO_WRITE_ERROR;					// A problem occurs in WriteFile
	}
	
	return FILEIO_NOERROR;	// no problem occurs
}



//////////////////////////////////////////////////////////////////////
// file function(s)
//////////////////////////////////////////////////////////////////////
INT32 FileIO_PCM_GetPosition(void *pInstance, INT32 *CurrentPosition)
{
	FileIO_PCM_Struct *pInst = (FileIO_PCM_Struct *)pInstance;

	// use standard function to get position and return if an error occured
	INT32 size = ftell(pInst->file);
	if (size == -1)
	{
		*CurrentPosition = 0;
		return FILEIO_CANNOT_GET_POSITION;
	}
	else
	{
		if (pInst->NumberOfChannels == 1) *CurrentPosition = size / 2;
		else *CurrentPosition = size / 4;
		return FILEIO_NOERROR;
	}
}

INT32 FileIO_PCM_GetFileLength(void *pInstance, INT32 *FileSize)
{
	INT32 result;
	INT32 CurrentPosition = 0, size;

	FileIO_PCM_Struct *pInst = (FileIO_PCM_Struct *)pInstance;

	// Get the current position in the file and return if impossible
	CurrentPosition = ftell(pInst->file);
	if (CurrentPosition == -1) return FILEIO_CANNOT_GET_POSITION;

	
	// Seek to the end of file
	result = fseek(pInst->file, 0, SEEK_END);
	if (result != 0) return FILEIO_PROBLEM_WHILE_SEEKING;
	
	// Get the position of the end of file
	size = ftell(pInst->file);
	if (size == -1) return FILEIO_CANNOT_GET_POSITION;


	//  Seek to the position where the file was before
	result = fseek(pInst->file, CurrentPosition, SEEK_SET);
	if (result != 0) return FILEIO_PROBLEM_WHILE_SEEKING;


	if (pInst->NumberOfChannels == 1) *FileSize = size / 2;
	else *FileSize = size / 4;


	return FILEIO_NOERROR;
}



//////////////////////////////////////////////////////////////////////
// control function(s)
//////////////////////////////////////////////////////////////////////
INT32 FileIO_PCM_SetLoopingMode(void *pInstance, INT32 Looping)
{
	FileIO_PCM_Struct *pInst = (FileIO_PCM_Struct *)pInstance;
	pInst->Loop = Looping;

	return FILEIO_NOERROR;
}



//////////////////////////////////////////////////////////////////////
// "text" function(s)
//////////////////////////////////////////////////////////////////////
void FileIO_PCM_GetLastErrorString(CHAR *str, INT32 LastErrorCode, INT32 MaxNumberOfCharacters)
{
	if ((LastErrorCode & FILEIO_BASE_ERR) == FILEIO_BASE_ERR)		// check if the error is a FileIO error
	{
		if ((LastErrorCode & FILEIO_GEN_BASE_ERR) == FILEIO_GEN_BASE_ERR)
		{
			FileIO_GetLastErrorString(str, LastErrorCode, MaxNumberOfCharacters);
			return;
		}
		else if ((LastErrorCode & FILEIO_PCM_BASE_ERR) == FILEIO_PCM_BASE_ERR)
		{
			// do nothing here, wait for the switch below
		}
		else
		{
			strncpy(str, "Error type not known", MaxNumberOfCharacters - 1);
		}
	}
	else															// this error is not a FileIO error
	{
		strncpy(str, "Not a FileIO error", MaxNumberOfCharacters - 1);
		return;
	}


	switch(LastErrorCode)
	{
		case FILEIO_PCM_NOERROR:
			strncpy(str, "PCM: No error occurred", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "PCM: Error not known", MaxNumberOfCharacters - 1);
			break;
	}
	return;
}



void FileIO_PCM_GetHelp(CHAR *str, INT32 LineNumber, INT32 MaxNumberOfCharacters)
{
	switch(LineNumber)
	{
		case 0:
			strncpy(str, "PCM files are read by these functions.", MaxNumberOfCharacters - 1);
			break;
		case 1:
			strncpy(str, "PCM files are raw data encoded in Little Endian format i.e. LSB,MSB", MaxNumberOfCharacters - 1);
			break;
		case 2:
			strncpy(str, "There's no support for 8 bits files and for more than 4GB files.", MaxNumberOfCharacters - 1);
			break;
		default:
			strncpy(str, "PCM file functions", MaxNumberOfCharacters - 1);
	}
}
